const test = require('../sprut-check/index');
const {
    configure,
    computed,
    value: reactive,
    callback,
    asyncWatcher,
    waiting,
    allIsWaiting,
    someIsWaiting,
    spread,
    deferred,
    immediate,
    asDependency,
    Waiting
} = require('./index');

test('Тестирование автоконфигурации хранилища', ({ value, event, when }) => {
    const Store = value('Store', () => {
        class Store {
            constructor() {
                configure(this);
            }

            prop = 1;

            get computed() {
                return this.prop;
            }

            set computed(v) {
                this.prop = v;
            }
        }

        return Store;
    });

    const store = value('Экземпляр Store', () => new Store(), testValue => {
        testValue
            .mustBeEquals('Prop', v => [v.prop, 1])
            .mustBeEquals('Computed', v => [v.computed, 1])
            .mustBeFasterThan('', 1);
    });
    event('Change prop', 'Изменение через computed сеттер и вызов onChange экземпляра стора', ({ happend, event }) => {
        store.computed = 2;
        store.onChange('prop', happend);
    }, { timeout: 10 });
    when('Change prop', () => {
        value('Значение prop объекта store', () => store, testValue => {
            testValue
                .mustBeEquals('Prop', v => [ v.prop, 2 ])
                .mustBeEquals('Computed', v => [ v.computed, 2 ]);
        });
    });
});

test('Тестирование конфигурации хранилища', ({ value, event, when }) => {
    let globalVar, globalVar_2, globalVar_3;

    const Store = value('Store', () => {
        const config = {
            reactives: [ 'reactiveProp', 'reactiveProp_2' ],
            reactive: {
                configReactiveProp: {
                    default: 'value'
                },
                immediateProp: {
                    default: 'value',
                    immediate: true
                },
                factoryProp: {
                    default: () => new EventTarget(),
                    isFactory: true
                }
            },
            computed: {
                computedProp: {
                    getter() {
                        return this.reactiveProp + 10;
                    },
                    setter(v) {
                        this.reactiveProp = v - 10;
                    }
                }
            },
            methods: {
                dependenciesAutoExecMethod: {
                    method() {
                        globalVar = this.computedProp;
                    },
                    autoExec: true,
                    dependencies() {
                        this.computedProp;
                    }
                },
                intiedStartAutoExecMethod: {
                    method() {
                        globalVar_2 = this.reactiveProp;
                    },
                    autoExec: true,
                    initedStart: true
                },
                callbackDependenciesAutoExecMethod: {
                    method() {
                        globalVar_3 = this.reactiveProp;
                    },
                    dependencies() {
                        this.reactiveProp;
                    },
                    callback() {
                        globalVar_3 = this.reactiveProp;
                    },
                }
            }
        };

        class Store {
            storeConfig = config;

            constructor() {
                configure(this);
            }

            reactiveProp = 0;

            reactiveProp_2 = 'value';

            notReactiveProp = 1;
        }

        return Store;
    });

    const store = value('Экземпляр Store', () => new Store(), testValue => {
        testValue
            .mustBeEquals('reactiveProp', v => [ v.reactiveProp, 0 ])
            .mustBeEquals('reactiveProp_2', v => [ v.reactiveProp_2, 'value' ])
            .mustBeEquals('notReactiveProp', v => [ v.notReactiveProp, 1 ])
            .mustBeEquals('configReactiveProp', v => [ v.configReactiveProp, 'value' ])
            .mustBeEquals('immediateProp', v => [ v.immediateProp, 'value' ])
            .mustBeTypeof('factoryProp', v => [ v.factoryProp, [ EventTarget ] ])
            .mustBeEquals('computedProp', v => [ v.computedProp, 10 ]);
    });

    value('Change var 1 into dependenciesAutoExecMethod', () => globalVar, testValue => {
        testValue.mustBeEquals('Computed prop', undefined);
    });
    value('Change var 2 into intiedStartAutoExecMethod', () => globalVar_2, testValue => {
        testValue.mustBeEquals('Computed prop', 0);
    });
    value('Change var 3 into dependenciesAutoExecMethod', () => globalVar, testValue => {
        testValue.mustBeEquals('Computed prop', undefined);
    });

    event('Change props', 'Изменения значений в сторе и тест автозапусков', ({ happend, event }) => {
        store.onChange('reactiveProp', happend);

        store.reactiveProp = 10;
    }).event('Change immediateProp', 'Изменение immediate свойства', ({ happend, event }) => {
        store.immediateProp = 'new value';

        value('Экземпляр Store', () => store, testValue => {
            testValue
                .mustBeEquals('immediateProp', v => [ v.immediateProp, 'new value' ]);
        });

        happend();
    });
    when('Change props', () => {
        value('Экземпляр Store', () => store, testValue => {
            testValue
                .mustBeEquals('reactiveProp', v => [ v.reactiveProp, 10 ])
                .mustBeEquals('computedProp', v => [ v.computedProp, 20 ]);
        });

        value('Change var 1 into dependenciesAutoExecMethod', () => globalVar, testValue => {
            testValue.mustBeEquals('Computed prop', 20);
        });
        value('Change var 2 into intiedStartAutoExecMethod', () => globalVar_2, testValue => {
            testValue.mustBeEquals('Computed prop', 10);
        });
        value('Change var 3 into dependenciesAutoExecMethod', () => globalVar, testValue => {
            testValue.mustBeEquals('Computed prop', 20);
        });
    });
});

test('Тестирование функциональных методов', ({ value, event, when }) => {
    const [ get_1, set_1 ] = value('Reactive value 1', () => reactive(3), testValue => {
        testValue.mustBeEquals('', (reactive) => [ reactive[0](), 3 ]);
    });
    const [ get_2, set_2 ] = value('Reactive value 2', () => reactive(20), testValue => {
        testValue.mustBeEquals('', (reactive) => [ reactive[0](), 20 ]);
    });
    const [ getComputed, setComputed ] = value('Computed value', () => computed(() => get_1() + get_2(), (v) => { set_1(v / 2); set_2(v / 2) }), testValue => {
        testValue.mustBeEquals('', (computed) => [ computed[0](), 23 ]);
    });

    event('Global update', 'Применение значения в computed и событие по коллбэку callback', ({ happend }) => {
        callback(() => {}, () => getComputed(), happend);
        setComputed(30);
    });
    when('Global update', () => {
        value('Reactive value 1', () => get_1(), testValue => {
            testValue.mustBeEquals('', 15);
        });
        value('Reactive value 2', () => get_2(), testValue => {
            testValue.mustBeEquals('', 15);
        });
        value('Computed value', () => getComputed(), testValue => {
            testValue.mustBeEquals('', 30);
        });
    });
});

test('Тестирование immediate и deferred в функциональных методах', ({ value, event, when }) => {
    const [ getValue, setValue ] = value('Reactive value', () => reactive(1), testValue => {
        testValue.mustBeEquals('', (reactive) => [ reactive[0](), 1 ]);
    });

    event('Deferred', 'Применение значения в value и событие по коллбэку callback', ({ happend }) => {
        const [_, off] = callback(() => {}, () => getValue(), happend);
        setValue(deferred(2));
        value('Reactive value after deferred set', () => getValue(), testValue => {
            testValue.mustBeEquals('must be prev value', 1);
        });
    }).event('Immediate', 'Применение значения в value и событие по коллбэку callback', ({ happend }) => {
        setValue(immediate(5));
        value('Reactive value after immediate set', () => getValue(), testValue => {
            testValue.mustBeEquals('must be new value', 5);
        });
        happend();
    });
    when('Deferred', () => {
        value('Reactive value deferred change', () => getValue(), testValue => {
            testValue.mustBeEquals('New value', 2);
        });
    });
});

test('Тестирование ожидания', ({ value, event, when }) => {
    const [ get_1, set_1 ] = value('Reactive value 1', () => reactive(3), testValue => {
        testValue.mustBeEquals('', (reactive) => [ reactive[0](), 3 ]);
    });
    const [ getComputed, setComputed ] = value(
        'Computed value',
        () => computed(
            () => {
                const res = get_1();

                return waiting((put) => {
                    setTimeout(() => put(res), 500);
                });
            },
            (v) => set_1(v)
        ), testValue => {
            testValue.mustBeEquals('', (computed) => [ computed[0](), Waiting ]);
        }
    );

    event('Reactive update', 'Применение значения в computed и событие по коллбэку callback', ({ happend }) => {
        callback(() => {}, () => getComputed(), happend);
        setComputed(1);
    }).event('Reactive update after waiting', 'Применение значения в computed и событие по коллбэку callback', ({ happend }) => {
        callback(() => {}, () => getComputed(), happend);
    }, { timeout: 1000 });
    when('Reactive update', () => {
        value('Computed value', () => getComputed(), testValue => {
            testValue.mustBeEquals('', Waiting);
        });
    });
    when('Reactive update after waiting', () => {
        value('Computed value', () => getComputed(), testValue => {
            testValue.mustBeEquals('', 1);
        });
    });
});

test('Тестирование spread', ({ value, event, when }) => {
    const Store = value('Store', () => {
        class Store {
            constructor() {
                configure(this);
            }

            prop = 12;

            get computed() {
                return this.prop + 10;
            }

            set computed(v) {
                this.prop = v;
            }

            method() {
                console.log('Something');
            }
        }

        return Store;
    });

    const store = value('Экземпляр Store', () => new Store());
    value('Чистый объект spread', () => spread(store), testValue => {
        testValue.mustBeEquals('Предполагаемый объект', { prop: 12, computed: 22 });
    });
});