let isStarted = false;
const tests = [];
const Operations = {
    equals: 'equals',
    typeof: 'typeof',
    execTest: 'execTest',
    happendAction: 'happendAction',
    execEvent: 'execEvent',
    execActionWhenEvent: 'execActionWhenEvent',
    mustBe: 'mustBe',
    eventFasterThan: 'eventFasterThan',
    valueFasterThan: 'valueFasterThan',
    eventEarlierThan: 'eventEarlierThan',
    execValue: 'execValue',
    execValueTest: 'execValueTest'
};
const FailedReason = {
    execValueError: 'execValueError'
};

const testInfo = [];

function test(testName, actions) {
    if (!isStarted) {
        setTimeout(() => {
            execTests();
        });
        isStarted = true;
    }

    const results = [];
    const events = [];
    const eventActions = {};
    let startTime;

    const testAction = () => {
        const testInfo = { results, testName, events, testEnded: () => events.reduce((res, e) => res && (!!e.happendTime || e.isTimeout), true) };

        startTime = testInfo.startTime = new Date().getTime();

        try {
            actions({value, event, when });
            results.push({
                time: new Date().getTime() - startTime,
                operation: Operations.execTest,
                result: true,
            })
        } catch(error) {
            results.push({
                time: new Date().getTime() - startTime,
                operation: Operations.execTest,
                error,
                result: false,
            });
        }

        return testInfo;
    }

    if (!isStarted) {
        tests.push(testAction);
    } else {
        testInfo.push(testAction());
    }

    function value(name, v, testing) {
        let catchResult;
        const valueStartTime = new Date().getTime();

        try {
            v = v();
        } catch(error) {
            catchResult = {
                time: new Date().getTime(),
                valueName: name,
                operation: Operations.execValue,
                error,
                result: false
            };
            results.push(catchResult);
        }

        const testObject = {
            mustBeEquals(_name, _v, symbols) {
                if (_v instanceof Function) {
                    var [ val, _v ] = _v(v);
                } else {
                    var val = v;
                }
                if (catchResult) {
                    results.push({
                        time: new Date().getTime(),
                        valueName: name,
                        operation: Operations.equals,
                        result: false,
                        error: catchResult.error,
                        reason: FailedReason.execValueError,
                        reasonText: catchResult.error.message
                    });

                    return testObject;
                }

                const time = new Date().getTime();

                results.push({
                    time,
                    valueName: name,
                    value: val,
                    compareName: _name,
                    compareValue: _v,
                    operation: Operations.equals,
                    result: isEquals(val, _v, symbols)
                });

                return testObject;
            },
            mustBeTypeof(description, types, isStrictFields) {
                if (types instanceof Function) {
                    var [ val, types ] = types(v);
                } else {
                    var val = v;
                }
                if (catchResult) {
                    results.push({
                        time: new Date().getTime(),
                        valueName: name,
                        operation: Operations.typeof,
                        result: false,
                        description,
                        error: catchResult.error,
                        reason: FailedReason.execValueError,
                        reasonText: catchResult.error.message
                    });

                    return testObject;
                }

                const time = new Date().getTime();
                const result = validateTypes(val, types, isStrictFields);

                results.push({
                    time,
                    valueName: name,
                    value: val,
                    types,
                    description,
                    operation: Operations.typeof,
                    result
                });

                return testObject;
            },
            mustBeFasterThan(description, milliseconds) {
                if (catchResult) {
                    results.push({
                        time: new Date().getTime(),
                        valueName: name,
                        operation: Operations.valueFasterThan,
                        result: false,
                        description,
                        error: catchResult.error,
                        reason: FailedReason.execValueError,
                        reasonText: catchResult.error.message
                    });

                    return testObject;
                }

                const time = new Date().getTime();

                results.push({
                    time,
                    valueStartTime,
                    valueName: name,
                    recommendedTime: milliseconds,
                    result: time - valueStartTime <= milliseconds,
                    description,
                    event,
                    operation: Operations.valueFasterThan
                });

                return testObject;
            },
            mustBeCatched(description, message, Error) {
                if (catchResult && (Error ? catchResult.error instanceof Error : true)) {
                    catchResult.description = description;
                    catchResult.result = (message ? catchResult.error.message === message : true) && (Error ? catchResult.error instanceof Error : true);
                }
            },
            mustBe() {
                
            },
        };

        if (testing instanceof Function) {
            try {
                testing(testObject)
            } catch(error) {
                results.push({
                    time: new Date().getTime(),
                    operation: Operations.execValueTest,
                    error,
                    result: false
                });

                return;
            }
        }

        return catchResult ? null : v;
    }

    function allEvents(...events) {
        const resolved = [];

        events.forEach(e => {
            e.events('', '')
        });
    }

    function event(name, description, actions, config = {}, chainData) {
        const { timeout, noLogs } = config;
        const eventObj = {
            name,
            description,
            actions: [],
            noLogs
        };
        const chainEvents = [];

        events.push(eventObj);

        const eventTests = {
            mustBeFasterThan(milliseconds) {
                const time = new Date().getTime();

                results.push({
                    time,
                    recommendedTime: milliseconds,
                    event: eventObj,
                    operation: Operations.eventFasterThan
                });

                return eventTests;
            },
            mustBeEarlierThan(milliseconds) {
                const time = new Date().getTime();

                results.push({
                    time,
                    recommendedTime: milliseconds,
                    event: eventObj,
                    operation: Operations.eventEarlierThan
                });

                return eventTests;
            }
        };

        try {
            eventObj.startTime = new Date().getTime();
            actions({ happend: (withResult, data) => {
                setTimeout(() => {
                    if (true || !eventObj.happendTime) {
                        eventObj.happendTime = new Date().getTime();
                        eventObj.withResult = withResult;
                        eventObj.data = data;

                        results.push({
                            time: new Date().getTime(),
                            event: eventObj,
                            operation: Operations.execEvent,
                            result: true
                        });

                        eventActions[name]?.forEach(action => {
                            try {
                                action({ eventData: event.data, testEvent: eventTests, event });

                                results.push({
                                    time: new Date().getTime(),
                                    event: eventObj,
                                    operation: Operations.execActionWhenEvent,
                                    result: true
                                });
                            } catch(error) {
                                results.push({
                                    time: new Date().getTime(),
                                    event: eventObj,
                                    operation: Operations.execActionWhenEvent,
                                    error,
                                    result: false
                                });
                            }
                        });

                        const chainEv = chainEvents[0];

                        chainEvents.shift();
                        chainEv ? chainEv() : undefined;
                    }
                });
            }, event: eventTests, chainData });
            if (timeout) {
                setTimeout(() => {
                    if (!eventObj.happendTime) {
                        eventObj.isTimeout = true;
                        results.push({
                            time: new Date().getTime(),
                            event: eventObj,
                            operation: Operations.execEvent,
                            error: { message: "event timeout failed", name: 'Error'},
                            result: false
                        });
                    }
                }, timeout);
            }
        } catch(error) {
            results.push({
                time: new Date().getTime(),
                event: eventObj,
                operation: Operations.execEvent,
                error,
                result: false
            });
        }

        return { event: function ev(name, description, actions, config) { chainEvents.push(() => event(name, description, actions, config, eventObj.data)); return { event: ev } }};
    }

    function when(eventName, actions) {
        if (!eventActions[eventName]) {
            eventActions[eventName] = [];
        }
        eventActions[eventName].push(actions);
    }
}

function execTests() {
    tests.forEach(test => testInfo.push(test()));
    waitingTests();
}

function waitingTests() {
    let finished = true;

    for(let info of testInfo) {
        finished = finished && info.testEnded();
    }

    if (!finished) {
        setTimeout(waitingTests);
    } else {
        resolveAndLogTests();
    }
}

function resolveAndLogTests() {
    const tests = [];

    testInfo.forEach(test => {
        const testResult = {};
        let results = testResult.results = [];
        testResult.testName = test.testName;
        tests.push(testResult);

        test.results.forEach(result => {
            if (result.result === undefined) {
                if (result.operation === Operations.eventFasterThan) {
                    result.resultTime = result.event.happendTime - result.event.startTime;
                    result.result = result.resultTime <= result.recommendedTime;
                } else if (result.operation === Operations.eventEarlierThan) {
                    result.resultTime = result.event.happendTime - test.startTime;
                    result.result = result.resultTime <= result.recommendedTime;
                }
            }
        });

        testResult.result = test.results.reduce((res, result) => res && result.result, true);

        test.results.forEach(result => {
            if (!result.result && 'reason' in result) {
                if (result.reason === FailedReason.execValueError) {
                    results.push({ time: result.time - test.startTime, description: `Test operation <${result.operation}> cannot be executed with error <${result.reasonText}>`, tooltip: result.error.stack, result: result.result });
                    return;
                } else {
                    results.push({ time: result.time - test.startTime, description: `Test operation <${result.operation}> cannot be executed for reason ${result.reason}`, tooltip: result.error.stack, result: result.result });
                    return;
                }
            }

            switch (result.operation) {
                case Operations.equals: {
                    results.push({ time: result.time - test.startTime, description: `Value "${result.valueName}" = ${result.value?.toString()} is${result.result ? ' ' : ' not '}equals to value ${result.compareName} = ${result.compareValue?.toString()}`, tooltip: `${typesToString([result.value])}${result.result ? ' ' : ' not '}equals ${typesToString([result.compareValue])}`, result: result.result });
                    break;
                }
                case Operations.typeof: {
                    if (result.result) {
                        results.push({ time: result.time - test.startTime, description: `Value "${result.valueName}" is type of ${typesToString(result.types)}`, result: result.result });
                    } else {
                        results.push({ time: result.time - test.startTime, description: `Value "${result.valueName}" must be type of ${typesToString(result.types)}`, result: result.result });
                    }
                    break;
                }
                case Operations.execValue: {
                    if (!result.result) {
                        results.push({ time: result.time - test.startTime, description: `Value "${result.valueName}" executed with ${result.error}`, tooltip: result.error.stack, result: result.result });
                    } else {
                        results.push({ time: result.time - test.startTime, description: `Value "${result.valueName}" catched with ${result.error}`, tooltip: result.error.stack, result: result.result });
                    }
                    break;
                }
                case Operations.execActionWhenEvent: {
                    if (result.result) {
                        results.push({ time: result.time - test.startTime, description: `Happend event action "${result.event.name}"(${result.event.description}) with result ${result.event.withResult instanceof Object ? '(hover to details)' : '"' + result.event.withResult + '"'}`, tooltip: result.event.withResult instanceof Object ? typesToString([result.event.withResult]) : '', result: result.result });
                    } else {
                        results.push({ time: result.time - test.startTime, description: `Event action "${result.event.name}"(${result.event.description}) catched with "${result.error.message}"`, tooltip: result.error.stack, result: result.result });
                    }
                    break;
                }
                case Operations.execEvent: {
                    if (!result.event.noLogs) {
                        if (result.result) {
                            results.push({ time: result.time - test.startTime, description: `Happend event "${result.event.name}"(${result.event.description}) with result ${result.event.withResult instanceof Object ? '(hover to details)' : '"' + result.event.withResult + '"'}`, tooltip: result.event.withResult instanceof Object ? typesToString([result.event.withResult]) : '', result: result.result });
                        } else {
                            results.push({ time: result.time - test.startTime, description: `Event "${result.event.name}"(${result.event.description}) catched with "${result.error.message}"`, tooltip: result.error.stack, result: result.result });
                        }
                    }
                    break;
                }
                case Operations.valueFasterThan: {
                    if (result.result) {
                        results.push({ time: result.time - test.startTime, description: `Value "${result.valueName}"(${result.description}) must be faster than ${(result.recommendedTime / 1000).toFixed(3)}s. Result: ${((result.time - result.valueStartTime) / 1000).toFixed(3)}s`, result: result.result });
                    } else {
                        results.push({ time: result.time - test.startTime, description: `Value "${result.valueName}"(${result.description}) must be faster than ${(result.recommendedTime / 1000).toFixed(3)}s. Result: ${((result.time - result.valueStartTime) / 1000).toFixed(3)}s`, result: result.result });
                    }
                    break;
                }
                case Operations.eventFasterThan: {
                    if (result.result) {
                        results.push({ time: result.time - test.startTime, description: `Value "${result.event.name}"(${result.event.description}) must be faster than ${(result.recommendedTime / 1000).toFixed(3)}s. Result: ${((result.event.happendTime - result.event.startTime) / 1000).toFixed(3)}s`, result: result.result });
                    } else {
                        results.push({ time: result.time - test.startTime, description: `Value "${result.event.name}"(${result.event.description}) must be faster than ${(result.recommendedTime / 1000).toFixed(3)}s. Result: ${((result.event.happendTime - result.event.startTime) / 1000).toFixed(3)}s`, result: result.result });
                    }
                    break;
                }
                default: {
                    results.push({ time: result.time, description: result.operation, result: result.result });
                    break;
                }
            }
        });
    });

    const path = require('path');
    const testView = require('fs').readFileSync(path.resolve(__dirname, './testViewer.html')).toString();
    const server = require('http').createServer((req, res) => {
        res.end(testView.replace(/'@content'/g, JSON.stringify(tests)));
        req.on('end', () => {
            setTimeout(() => {
                server.closeAllConnections();
                server.close();
            }, 1000);
        });
    });
    let port = 3001;
    server.listen(port, () => require('child_process').exec(`start chrome http://localhost:${port}`));
    server.on('error', (err) => {
        if (/address already in use/.test(err.message)) {
            port++;
            server.listen(port);
        }
    });
    //console.log(testInfo.map(t => t.results));
}

function typesToString(types = []) {
    let result = [];

    types.forEach(type => {
        if (type instanceof Function) {
            result.push(`<${type.prototype.constructor.name}>`);
        } else if (type instanceof Object) {
            result.push(`<${JSON.stringify(type)}>`);
        } else if (type === 'string') {
            result.push(`"${type}"`)
        } else if (type === 'number') {
            result.push(`${type}`)
        } else if (typeof type === 'string') {
            result.push(`"${type}"`)
        } else if (type === undefined) {
            result.push("undefined")
        } else if (type === null) {
            result.push("null")
        } else if (type === NaN) {
            result.push("NaN")
        } else {
            result.push(type?.toString())
        }
    });

    return `( ${result.join(' | ')} )`;
}

function validateTypes(value, types, isStrictFields) {
    for(let type of types) {
        if(
            (type instanceof Function && value instanceof type)
            || typeof value === type
            || (type instanceof Object && isContainsProps(value, type, isStrictFields))
            || (!(type instanceof Object) && value === type)
        ) {
            return true;
        }
    }

    return false;
}

function isContainsProps(value, type, isStrictFields) {
    if (isStrictFields && Object.keys(value).length !== Object.keys(type).length) {
        return false;
    }

    for(prop in type) {
        if (!(prop in value && validateTypes(value[prop], [type[prop]]))) {
            return false;
        }
    }

    return true;
}

function isEquals(o_1, o_2, symbols) {
    if (o_1 === o_2) {
        return true;
    }

    if (o_1 instanceof Array && o_2 instanceof Array) {
        if (o_1.length !== o_2.length) {
            return false;
        }

        for(let i in o_1) {
            if (!isEquals(o_1[i], o_2[i], symbols)) {
                return false;
            }
        }

        return isEqualsSymbols(o_1, o_2, symbols);
    }

    if (o_1 instanceof Object && o_2 instanceof Object) {
        if (Object.keys(o_1).length !== Object.keys(o_2).length) {
            return false;
        }

        for(let prop in o_1) {
            if (!isEquals(o_1[prop], o_2[prop], symbols)) {
                return false;
            }
        }

        return isEqualsSymbols(o_1, o_2, symbols);
    }

    return false;
}

function isEqualsSymbols(o_1, o_2, symbols = []) {
    for(let s of symbols) {
        if (s in o_1 || s in o_2 && (s in o_1 && s in o_2)) {
            if (!isEquals(o_1[s], o_2[s], symbols)) {
                return false;
            }
        }
    }

    return true;
}

function isContains(o_1, o_2, symbols) {
    if (o_1 === o_2) {
        return true;
    }

    if (o_1 instanceof Array && o_2 instanceof Array) {
        if (o_1.length < o_2.length) {
            return false;
        }

        for(let i in o_2) {
            if (!isContains(o_1[i], o_2[i], symbols)) {
                return false;
            }
        }

        return isContainsSymbols(o_1, o_2, symbols);
    }

    if (o_1 instanceof Object && o_2 instanceof Object) {
        if (Object.keys(o_1).length < Object.keys(o_2).length) {
            return false;
        }

        for(let prop in o_2) {
            if (!isContains(o_1[prop], o_2[prop], symbols)) {
                return false;
            }
        }

        return isContainsSymbols(o_1, o_2, symbols);
    }

    return false;
}

function isContainsSymbols(o_1, o_2, symbols = []) {
    for(let s of symbols) {
        if (s in o_1 && s in o_2) {
            if (!isContains(o_1[s], o_2[s], symbols)) {
                return false;
            }
        } else if (s in o_2 && !(s in o_1)) {
            return false;
        }
    }

    return true;
}

function isCrossedEquals(o_1, o_2, symbols) {
    if (o_1 === o_2) {
        return true;
    }

    if (o_1 instanceof Array && o_2 instanceof Array) {
        for(let i in o_2) {
            if (i in o_1 && i in o_2) {
                if (!isCrossedEquals(o_1[i], o_2[i], symbols)) {
                    return false;
                }
            }
        }

        return isCrossedSymbols(o_1, o_2, symbols);
    }

    if (o_1 instanceof Object && o_2 instanceof Object) {
        for(let prop in o_2) {
            if (prop in o_1 && prop in o_2) {
                if (!isCrossedEquals(o_1[prop], o_2[prop], symbols)) {
                    return false;
                }
            }
        }

        return isCrossedSymbols(o_1, o_2, symbols);
    }

    return false;
}

function isCrossedSymbols(o_1, o_2, symbols = []) {
    for(let s of symbols) {
        if (s in o_1 && s in o_2) {
            if (!isCrossedEquals(o_1[s], o_2[s], symbols)) {
                return false;
            }
        }
    }

    return true;
}

module.exports = test;